# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving Tasks."""

import abc
import datetime as dt
import hashlib
import shlex
import tarfile
from io import BytesIO
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Any, ClassVar, cast, override
from unittest import mock

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.client.debusine import Debusine
from debusine.client.models import (
    ArtifactResponse,
    FileResponse,
    FilesResponseType,
)
from debusine.tasks import BaseExternalTask, BaseTaskWithExecutor, inputs
from debusine.tasks.executors import ExecutorInterface, UnshareInstance
from debusine.tasks.inputs.mock_resolver import MockFieldResolver
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseDynamicTaskDataWithExecutor,
    BaseTaskData,
    BaseTaskDataWithExecutor,
)
from debusine.test.base import TestCase
from debusine.test.test_utils import create_artifact_response
from debusine.utils import extract_generic_type_arguments


class CurrentTaskAccessor[ValueType]:
    """Define a class member that depends on the current task being set."""

    name: str

    def __set_name__(self, owner: type["TaskTestCase[Any]"], name: str) -> None:
        """Set the field name at class construction."""
        self.name = name

    def __get__(
        self,
        obj: "TaskTestCase[Any]",
        _: "type[TaskTestCase[Any]] | None" = None,
    ) -> ValueType:
        """Initialize the current task the first time the field is accessed."""
        obj.task_set_current(obj.task_create())
        return cast(ValueType, obj.__dict__[self.name])


class TaskTestCase[TaskClass](TestCase, abc.ABC):
    """
    Base test case for all task types.

    The task being tested is instantiated the first time it is accessed, so you
    can edit self.task_data before access to customize the task data for the
    current test.
    """

    #: Common task data for all tests
    SAMPLE_TASK_DATA: ClassVar[dict[str, Any]] = {}
    #: Test-specific task data. It is initialized with SAMPLE_TASK_DATA and can
    #: be edited freely
    task_data: dict[str, Any]

    #: Class for the task being tested
    task_class: type[TaskClass]

    #: Tasks instantiated by the current test
    tasks: list[TaskClass]

    task = CurrentTaskAccessor[TaskClass]()

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Compute cls.task_class."""
        super().__init_subclass__(**kwargs)

        # The task class used for this test suite, computed by introspecting
        # the type argument used to specialize this generic class.
        [cls.task_class] = extract_generic_type_arguments(cls, TaskTestCase)

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data = self.SAMPLE_TASK_DATA.copy()
        self.tasks = []

    @override
    def tearDown(self) -> None:
        """Delete debug log files directory if it exists."""
        for task in self.tasks:
            self.task_teardown(task)
        super().tearDown()

    @abc.abstractmethod
    def task_create(self, **kwargs: Any) -> TaskClass:
        """
        Instantiate a task using the current task data.

        :param kwargs: if any keyword arguments are provided, they replace the
          corresponding values in ``self.task_data``.
        """

    def task_manage(self, task: TaskClass) -> None:
        """Register a task for task_setup and task_teardown."""
        self.task_setup(task)
        self.tasks.append(task)

    def task_setup(self, task: TaskClass) -> None:
        """Instantiate the task to test."""

    def task_teardown(self, task: TaskClass) -> None:
        """Perform cleanup at the end of the lifetime of the current task."""

    def task_reset_current(self) -> None:
        """Remove the current task, if already instantiated."""
        self.__dict__.pop("task", None)

    def task_set_current(self, task: TaskClass) -> None:
        """Set the given task as self.task."""
        if task not in self.tasks:
            self.task_manage(task)
        self.__dict__["task"] = task

    @abc.abstractmethod
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        """Invoke compute_scheduler_tags."""

    def build_expected_tag_candidates(
        self, **kwargs: set[str]
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        """
        Build sets of candidate provided and required tags.

        Extra tags can be added by providing sets of tags to arguments named
        ``{provide|require}_{provenance}``
        """
        expected_provided = ttags.TaskProvidedTags()
        for name, tags in kwargs.items():
            if name.startswith("provide_"):
                expected_provided.add(
                    ttags.ProvenanceProvided(name.removeprefix("provide_")),
                    tags,
                )

        expected_required = wtags.TaskRequiredTags()
        for name, tags in kwargs.items():
            if name.startswith("require_"):
                expected_required.add(
                    wtags.ProvenanceRequired(name.removeprefix("require_")),
                    tags,
                )

        return expected_provided, expected_required

    def assertComputesSchedulerTags(self, **kwargs: set[str]) -> None:
        """
        Compute scheduler tags and check that they match the expected values.

        To avoid having to repeat common tags for all tests, this generates two
        expected tag sets with basic values, and lets the user add to them.

        You can specify extra tags to add in arguments named
        ``{provide|require}_{provenance}`` with tag sets to add to the given
        expected tag set.
        """
        expected_provided, expected_required = (
            self.build_expected_tag_candidates(**kwargs)
        )
        provided, required = self.compute_scheduler_tags()
        assert provided.sets == expected_provided.sets
        assert required.sets == expected_required.sets


class ExternalTaskTestCase[TaskClass: BaseExternalTask[Any, Any]](
    TaskTestCase[TaskClass]
):
    """Base test case for external (worker and signing) tasks."""

    #: Field resolver for the test
    resolver: MockFieldResolver

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver = MockFieldResolver()

    @override
    def task_setup(self, task: TaskClass) -> None:
        super().task_setup(task)
        # The following values are set by the Worker
        task.work_request_id = 147
        task.workspace_name = "testing"
        task.working_directory = Path(self.create_temporary_directory())

    @override
    def task_create(self, **kwargs: Any) -> TaskClass:
        task_data = self.task_data.copy()
        task_data.update(kwargs)
        task = self.task_class(task_data)
        return task

    def get_subject(self) -> str | None:
        """Compute the subject for this task."""
        assert self.resolver is not None
        self.resolver.resolve_inputs(self.task, stage=inputs.Stage.PENDING)
        return self.task.get_subject()

    def get_configuration_context(self) -> str | None:
        """Compute the configuration context for this task."""
        assert self.resolver is not None
        self.resolver.resolve_inputs(self.task, stage=inputs.Stage.PENDING)
        return self.task.get_configuration_context()

    @override
    def compute_scheduler_tags(
        self, task: TaskClass | None = None
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        task = task or self.task
        self.resolver.resolve_inputs(task, stage=inputs.Stage.PENDING)
        return task.compute_scheduler_tags()

    # TODO: this currently cannot use self.task because it cannot be typed to
    # return the task's dynamic data type. Passing self.task is shorter to type
    # than casting or passing the dynamic data type
    def compute_dynamic_data[DTD: BaseDynamicTaskData](
        self, task: BaseExternalTask[Any, DTD]
    ) -> DTD:
        """Compute dynamic data for this task."""
        self.resolver.resolve_inputs(task)
        return task.compute_dynamic_data()

    def mock_debusine(self, task: TaskClass | None = None) -> mock.MagicMock:
        """Create a Debusine mock and configure self.task for it. Return it."""
        task = task or self.task
        debusine_mock = mock.create_autospec(spec=Debusine)
        assert isinstance(debusine_mock, mock.MagicMock)
        # Worker would set the server
        task.configure_server_access(debusine_mock)
        return debusine_mock


class TaskHelperMixin[TaskClass: BaseExternalTask[Any, Any]](
    ExternalTaskTestCase[TaskClass]
):
    """Helper mixin for Task tests."""

    def configure_task(
        self,
        task_data: dict[str, Any] | None = None,
        override: dict[str, Any] | None = None,
        remove: list[str] | None = None,
    ) -> TaskClass:
        """
        Run self.task.configure(task_data) with different inputs.

        Copy self.SAMPLE_TASK_DATA or task_data and modify it, call
        self.task.configure(modified_SAMPLE_TASK_DATA).

        :param task_data: if provided use this as a base dictionary. If None,
          use self.SAMPLE_TASK_DATA.
        :param override: dictionary with the keys to modify (only root keys,
          not recursive search) by its values. Modify or add them.
        :param remove: list of keys to remove.
        """
        if task_data is None:  # pragma: no cover
            task_data = self.SAMPLE_TASK_DATA.copy()
        if override:
            for key, value in override.items():
                task_data[key] = value
        if remove:
            for key in remove:
                task_data.pop(key, None)
        self.task_data = task_data

        # Reset the cached task if one was already built
        self.task_reset_current()
        # Call self.task to instantiate it, as some test methods call
        # configure_task as a way of instantiating a task with the given
        # arguments and see if it breaks
        return self.task

    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Invoke can_run_on."""
        self.resolver.resolve_inputs(self.task)
        return self.task.can_run_on(worker_metadata)

    def set_dynamic_data(self) -> None:
        """
        Compute and store the task's dynamic data.

        ``resolver`` must be set up in advance with enough information to
        allow computing dynamic data.
        """
        self.task.dynamic_data = self.compute_dynamic_data(self.task)

    def write_os_release(
        self, filename: Path, codename: str = "bookworm", version: int = 12
    ) -> dict[str, str]:
        """
        Write in directory / self.task._OS_RELEASE_FILE a release file.

        :return: dictionary with keys to values.
        """
        data = {
            "ID": "debian",
            "NAME": "Debian GNU/Linux",
            "PRETTY_NAME": f"Debian GNU/Linux {version} ({codename})",
            "VERSION": f"{version} ({codename})",
            "VERSION_CODENAME": codename,
            "VERSION_ID": str(version),
        }

        if version >= 13:
            del data["VERSION"]
            del data["VERSION_ID"]

        filename.write_text(
            "\n".join(f"{k}={shlex.quote(v)}" for k, v in data.items()) + "\n"
        )
        return data


class ExternalTaskHelperMixin[ExternalTaskClass: BaseExternalTask[Any, Any]](
    TaskHelperMixin[ExternalTaskClass]
):
    """Helper mixin for testing Tasks that run on external workers."""

    task_class: type[ExternalTaskClass]
    task: ExternalTaskClass

    def mock_image_download(
        self,
        debusine_mock: mock.MagicMock,
        create_image: bool = False,
        system_image: bool = False,
    ) -> ArtifactResponse:
        """
        Configure a fake ImageCache path and fake_system_*_artifact.

        The tarball itself isn't faked, unless create_image is set.

        If system_image is set, a system-image is faked instead of a tarball.
        """
        if system_image:
            artifact_response = self.fake_system_image_artifact()
        else:
            artifact_response = self.fake_system_tarball_artifact()
        debusine_mock.artifact_get.return_value = artifact_response

        def download_tarball(
            artifact_id: int,  # noqa: ARG001, U100
            filename: str,  # noqa: ARG001, U100
            destination: Path,
        ) -> None:
            with tarfile.open(destination, "w:xz") as f:
                tarinfo = tarfile.TarInfo("./")
                tarinfo.type = tarfile.DIRTYPE
                f.addfile(tarinfo)
                tarinfo = tarfile.TarInfo("./sbin")
                tarinfo.type = tarfile.DIRTYPE
                f.addfile(tarinfo)
                tarinfo = tarfile.TarInfo("./sbin/init")
                tarinfo.size = 3
                f.addfile(tarinfo, BytesIO(b"123"))

        def download_image(
            artifact_id: int,  # noqa: ARG001, U100
            filename: str,  # noqa: ARG001, U100
            destination: Path,
        ) -> None:
            with destination.open("wb") as f:
                f.write(b"qcow!")

        self.image_cache_path = Path(
            mkdtemp(prefix="debusine-testsuite-images-")
        )
        cast(TestCase, self).addCleanup(rmtree, self.image_cache_path)

        img_cache_patcher = mock.patch(
            "debusine.tasks.executors.images.ImageCache.image_cache_path",
            self.image_cache_path,
        )
        img_cache_patcher.start()
        cast(TestCase, self).addCleanup(img_cache_patcher.stop)

        patcher = mock.patch(
            "debusine.tasks.executors.images.ImageCache"
            "._download_image_artifact"
        )
        self.download_image_artifact_mock = patcher.start()
        cast(TestCase, self).addCleanup(patcher.stop)

        if create_image:
            if system_image:
                self.download_image_artifact_mock.side_effect = download_image
            else:
                self.download_image_artifact_mock.side_effect = download_tarball

        return artifact_response

    def patch_prepare_executor_instance(self) -> mock.MagicMock:
        """
        Patch self.task._prepare_executor_instance(), return its mock.

        Side effect of self.task._prepare_executor_instance(): set
        self.task.executor and self.task.executor_instance.
        """
        patcher = mock.patch.object(
            self.task, "_prepare_executor_instance", autospec=True
        )
        mocked = patcher.start()

        def mock_executor_executor_instance() -> None:
            self.task.executor = mock.MagicMock(spec=ExecutorInterface)
            self.task.executor_instance = mock.MagicMock(spec=UnshareInstance)
            self.task.executor_instance.run.return_value.returncode = 0

        mocked.side_effect = mock_executor_executor_instance
        cast(TestCase, self).addCleanup(patcher.stop)
        return mocked

    def fake_system_tarball_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:system-tarball."""
        return create_artifact_response(
            id=42,
            workspace="Testing",
            category=ArtifactCategory.SYSTEM_TARBALL,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "architecture": "amd64",
                "codename": "bookworm",
                "filename": "system.tar.xz",
                "vendor": "debian",
                "pkglist": {},
            },
            download_tar_gz_url="https://example.com/download-42/",
            files=FilesResponseType(
                {
                    "system.tar.xz": FileResponse(
                        size=4001,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/download-system.tar.xz",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_system_image_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:system-image."""
        return create_artifact_response(
            id=42,
            workspace="Testing",
            category=ArtifactCategory.SYSTEM_IMAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "architecture": "amd64",
                "codename": "bookworm",
                "filename": "image.qcow2",
                "vendor": "debian",
                "image_format": "qcow2",
                "filesystem": "ext4",
                "size": 2 * 1024 * 1024 * 1024,
                "boot_mechanism": "efi",
                "pkglist": {},
            },
            download_tar_gz_url="https://example.com/download-42/",
            files=FilesResponseType(
                {
                    "image.qcow2": FileResponse(
                        size=4001,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/download-image.qcow2",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_source_package_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:source-package."""
        return create_artifact_response(
            id=6,
            workspace="Testing",
            category=ArtifactCategory.SOURCE_PACKAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "name": "foo",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            download_tar_gz_url="https://example.com/download-6/",
            files=FilesResponseType(
                {
                    "foo_1.0-1.dsc": FileResponse(
                        size=10,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/foo_1.0-1.dsc",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_binary_package_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:binary-package."""
        return create_artifact_response(
            id=7,
            workspace="Testing",
            category=ArtifactCategory.BINARY_PACKAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "srcpkg_name": "foo",
                "srcpkg_version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
                "dsc_control_files": [],
            },
            download_tar_gz_url="https://example.com/download-7/",
            files=FilesResponseType(
                {
                    "foo_1.0-1_all.deb": FileResponse(
                        size=10,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/foo_1.0-1_all.deb",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_upload_artifact(
        self,
        source_package_name: str = "foo",
        architecture: str = "source",
        names: list[str] | None = None,
    ) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:upload."""
        if names is None:
            names = [
                "{source_package_name}_1.0.orig.tar.xz",
                "{source_package_name}_1.0-1.debian.tar.xz",
                "{source_package_name}_1.0-1.dsc",
            ]
        return create_artifact_response(
            id=8,
            workspace="Testing",
            category=ArtifactCategory.UPLOAD,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": architecture,
                    "Files": [{"name": name} for name in names],
                },
            },
            download_tar_gz_url="https://example.com/download-8/",
            files=FilesResponseType(
                {
                    name: FileResponse(
                        size=len(name),
                        checksums={
                            "sha256": hashlib.sha256(name.encode()).hexdigest()
                        },
                        type="file",
                        url=f"https://example.com/{name}",
                    )
                    for name in [
                        f"{source_package_name}_1.0-1_{architecture}.changes",
                        *names,
                    ]
                }
            ),
            files_to_upload=[],
        )


class SampleBaseExternalTask[
    TD: BaseTaskData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](
    BaseExternalTask[TD, DTD],
    metaclass=abc.ABCMeta,
):
    """Common test implementation of BaseExternalTask methods."""

    @override
    def get_subject(self) -> str | None:
        raise NotImplementedError()

    @override
    def compute_dynamic_data(self) -> DTD:
        """Resolve artifact lookups for this task."""
        return self.dynamic_task_data_type()

    @override
    def run(
        self,
        execute_directory: Path,  # noqa: ARG002, U100
    ) -> WorkRequestResults:
        raise NotImplementedError()


class SampleBaseTaskWithExecutor[
    TDE: BaseTaskDataWithExecutor,
    DTDE: BaseDynamicTaskDataWithExecutor,
](SampleBaseExternalTask[TDE, DTDE], BaseTaskWithExecutor[TDE, DTDE]):
    """Common test implementation of BaseExternalTask methods."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to collections in workspaces."""

from django.urls import path

import debusine.web.views.collection_relations as views

app_name = "collection_relations"

urlpatterns = [
    path(
        "",
        views.CollectionRelationListView.as_view(),
        name="list",
    ),
    path(
        "<int:pk>/delete/",
        views.CollectionRelationDeleteView.as_view(),
        name="delete",
    ),
    path(
        "create/",
        views.CollectionRelationCreateView.as_view(),
        name="create",
    ),
    path(
        "reorder-requires/",
        views.CollectionRelationReorderRequiresView.as_view(),
        name="reorder_requires",
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for management/utils.py."""

import datetime as dt
from typing import assert_never

from django.conf import settings
from django.core.management import CommandError

from debusine.db.models import Workspace
from debusine.server.management import management_utils
from debusine.test.django import TestCase


class ManagementUtilsTests(TestCase):
    """Tests for methods in management/management_utils.py."""

    def test_column_to_data(self) -> None:
        """Test Column.to_data."""
        col = management_utils.Column("foo", "Foo")
        for val in (None, 1, "str", dt.datetime.now(tz=dt.UTC)):
            with self.subTest(val=repr(val)):
                self.assertIs(col.to_data(val), val)

    def test_column_to_rich(self) -> None:
        """Test Column.to_rich."""
        col = management_utils.Column("foo", "Foo")
        d = dt.datetime.now(tz=dt.UTC)
        for val, expected in (
            (None, "-"),
            (1, "1"),
            ("str", "str"),
            (d, d.isoformat()),
        ):
            with self.subTest(val=repr(val)):
                self.assertEqual(col.to_rich(val), expected)

    def test_attrcolumn_to_data(self) -> None:
        """Test Column.to_data."""
        self.value = 42
        col = management_utils.AttrColumn("value", "Value", "value")
        self.assertIs(col.to_data(self), 42)

    def test_get_scope(self) -> None:
        """Test get_scope."""
        self.assertEqual(
            management_utils.get_scope(settings.DEBUSINE_DEFAULT_SCOPE),
            self.playground.get_default_scope(),
        )
        with self.assertRaisesRegex(
            CommandError, r"Scope 'test' not found"
        ) as exc:
            management_utils.get_scope("test")
        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_get_workspace(self) -> None:
        """Test get_workspace."""
        default_workspace = self.playground.get_default_workspace()
        scope = self.playground.get_or_create_scope("scope")
        workspace = self.playground.create_workspace(
            name="workspace", scope=scope
        )
        for name, require_scope, expected in (
            (settings.DEBUSINE_DEFAULT_WORKSPACE, False, default_workspace),
            (
                "workspace",
                False,
                f"Workspace 'workspace' not found in scope "
                f"'{settings.DEBUSINE_DEFAULT_SCOPE}'",
            ),
            (
                "workspace",
                True,
                "scope_workspace 'workspace' should be in the form "
                "'scopename/workspacename'",
            ),
            (
                f"{settings.DEBUSINE_DEFAULT_SCOPE}/"
                f"{settings.DEBUSINE_DEFAULT_WORKSPACE}",
                False,
                default_workspace,
            ),
            ("scope/workspace", False, workspace),
            (
                "does-not-exist/workspace",
                False,
                "Scope 'does-not-exist' not found",
            ),
            (
                f"{settings.DEBUSINE_DEFAULT_SCOPE}/workspace",
                False,
                f"Workspace 'workspace' not found in scope "
                f"'{settings.DEBUSINE_DEFAULT_SCOPE}'",
            ),
        ):
            with self.subTest(name=name):
                match expected:
                    case Workspace():
                        self.assertEqual(
                            management_utils.get_workspace(
                                name, require_scope=require_scope
                            ),
                            expected,
                        )
                    case str():
                        with self.assertRaisesRegex(
                            CommandError, expected
                        ) as exc:
                            management_utils.get_workspace(
                                name, require_scope=require_scope
                            )
                        self.assertEqual(exc.exception.returncode, 3)
                    case _ as unreachable:
                        assert_never(unreachable)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the SbuildValidatorMixin class."""

from pathlib import Path
from textwrap import dedent
from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.tasks.models import BaseTaskData
from debusine.tasks.sbuild_validator_mixin import SbuildValidatorMixin
from debusine.tasks.tests.helpers import SampleBaseExternalTask
from debusine.test import TestCase
from debusine.utils import calculate_hash


class TaskWithSbuildValidatorMixin(
    SbuildValidatorMixin, SampleBaseExternalTask[BaseTaskData]
):
    """Minimum class to test SbuildValidatorMixin."""

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Unused abstract method from BaseExternalTask."""
        raise NotImplementedError()


class SbuildValidatorMixinTests(TestCase):
    """Tests for SbuildValidatorMixin class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sbuild_validator_mixin = TaskWithSbuildValidatorMixin({})

    def test_check_directory_for_consistency_errors_no_errors(self) -> None:
        """check_directory_for_consistency_errors() return no errors."""
        sbuild_directory = self.create_temporary_directory()

        (file1 := sbuild_directory / "file1.deb").write_text("First file")
        (file2 := sbuild_directory / "file2.deb").write_text("Second file")

        # build_file does not cause any error
        (
            build_file := sbuild_directory
            / "hello_2.10-2_amd64-2023-05-08T10:04:40Z.build"
        ).write_text("the build file")

        # build_directory does not cause any error
        (sbuild_directory / "hello_2.10-2_amd64.build").symlink_to(build_file)

        changes_file = sbuild_directory / "hello.changes"

        self.write_dsc_file(sbuild_directory / "hello.dsc", [])

        self.write_changes_file(changes_file, [file1, file2])

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            [],
        )

    def test_check_directory_for_consistency_errors_no_changes_file(
        self,
    ) -> None:
        """check_directory_for_consistency_errors() handles missing .changes."""
        sbuild_directory = self.create_temporary_directory()

        self.sbuild_validator_mixin.check_directory_for_consistency_errors(
            sbuild_directory
        )

    def test_check_directory_for_consistency_errors_multiple_path_components(
        self,
    ) -> None:
        sbuild_directory = self.create_temporary_directory()
        changes_file = sbuild_directory / "hello.changes"
        changes_file.write_text(
            dedent("""\
                Format: 1.8
                Source: hello
                Version: 1.0
                Files:
                 0000 1 devel optional ../file
                Checksums-Sha1:
                 0000 1 ../file
                Checksums-Sha256:
                 0000 1 ../file
            """)
        )
        expected_errors = [
            "'hello.changes' has an entry in Files with more than one path"
            " component: '../file'"
        ]

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            expected_errors,
        )

    def test_check_directory_for_consistency_errors_missing_file(self) -> None:
        """check_directory_for_consistency_errors() return one file missing."""
        sbuild_directory = self.create_temporary_directory()

        (file1 := sbuild_directory / "file1.deb").write_text("First file")
        (file2 := sbuild_directory / "file2.deb").write_text("Second file")

        changes_file = sbuild_directory / "hello.changes"

        self.write_changes_file(changes_file, [file1, file2])

        file2.unlink()

        expected_errors = [
            f"File in .changes section {section} does not exist: 'file2.deb'"
            for section in ("Checksums-Sha1", "Checksums-Sha256", "Files")
        ]

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            expected_errors,
        )

    def test_check_directory_for_consistency_errors_files_in_dsc_no_error(
        self,
    ) -> None:
        """
        check_directory_for_consistency_errors(): no error.

        File is in .dsc and in the directory.
        """
        sbuild_directory = self.create_temporary_directory()

        (file_in_dsc := sbuild_directory / "a_file.txt").write_text("Some file")
        dsc_file = sbuild_directory / "hello.dsc"
        self.write_dsc_file(dsc_file, [file_in_dsc])

        changes_file = sbuild_directory / "hello.changes"
        self.write_changes_file(changes_file, [dsc_file])

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            [],
        )

    def test_check_directory_for_consistency_errors_bad_dsc_file(self) -> None:
        sbuild_directory = self.create_temporary_directory()

        (dsc_file := sbuild_directory / "hello.dsc").write_text("")

        changes_file = sbuild_directory / "hello.changes"
        self.write_changes_file(changes_file, [dsc_file])

        expected_errors = [f"{dsc_file.name!r} does not contain Source field"]

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            expected_errors,
        )

    def test_check_directory_for_consistency_errors_invalid_hash(self) -> None:
        """
        check_directory_for_consistency_errors() return an error.

        A file has an invalid hash.
        """
        sbuild_directory = self.create_temporary_directory()

        (file := sbuild_directory / "file1.deb").write_text("First file")
        expected_size = file.stat().st_size
        expected_sha256 = calculate_hash(file, "sha256").hex()
        expected_sha1 = calculate_hash(file, "sha1").hex()
        expected_md5 = calculate_hash(file, "md5").hex()

        changes_file = sbuild_directory / "hello.changes"

        self.write_changes_file(changes_file, [file])

        file.write_text("Something else")

        actual_size = file.stat().st_size
        actual_sha256 = calculate_hash(file, "sha256").hex()
        actual_sha1 = calculate_hash(file, "sha1").hex()
        actual_md5 = calculate_hash(file, "md5").hex()

        def format_error(
            hash_algo: str,
            filename: str,
            expected_hash: str,
            actual_hash: str,
            expected_size: int,
            actual_size: int,
        ) -> str:
            return (
                f'{hash_algo} for file {filename} does not match: '
                f'expected: {expected_hash} actual: {actual_hash}. '
                f'Size expected: {expected_size} actual: {actual_size}'
            )

        expected_errors = [
            format_error(
                "sha1",
                file.name,
                expected_sha1,
                actual_sha1,
                expected_size,
                actual_size,
            ),
            format_error(
                "sha256",
                file.name,
                expected_sha256,
                actual_sha256,
                expected_size,
                actual_size,
            ),
            format_error(
                "md5",
                file.name,
                expected_md5,
                actual_md5,
                expected_size,
                actual_size,
            ),
        ]

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                sbuild_directory
            ),
            expected_errors,
        )

    def test_check_directory_for_consistency_errors_extra_file(self) -> None:
        """
        check_directory_for_consistency_errors() return an error.

        There is an extra file in the directory.
        """
        build_directory = self.create_temporary_directory()
        changes_file = build_directory / "file.changes"

        (file1 := build_directory / "file1.deb").write_text("First file")

        self.write_changes_file(changes_file, [file1])

        (file2 := build_directory / "file2.deb").write_text("Second file")

        self.assertEqual(
            self.sbuild_validator_mixin.check_directory_for_consistency_errors(
                build_directory
            ),
            [
                "Files in directory not referenced in "
                f".changes or .dsc: ['{file2}']"
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the signing migrations."""

from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.test import TestCase

from debusine.assets import KeyPurpose
from debusine.signing.models import ProtectedKeyStorage


class MigrationTests(TestCase):
    """Test signing migrations."""

    def test_key_discriminated_union(self) -> None:
        """Keys are converted to the current layout for discriminated unions."""
        migrate_from = [("signing", "0001_initial")]
        migrate_to = [("signing", "0002_key_discriminated_union")]
        old_keys = [
            (
                "0" * 64,
                {
                    "storage": ProtectedKeyStorage.NACL,
                    "data": {"public_key": "pub1", "encrypted": "enc1"},
                },
                b"pub1",
            ),
            (
                "1" * 64,
                {
                    "storage": ProtectedKeyStorage.NACL,
                    "data": {"public_key": "pub2", "encrypted": "enc2"},
                },
                b"pub2",
            ),
        ]
        new_keys = [
            (
                "0" * 64,
                {
                    "storage": ProtectedKeyStorage.NACL,
                    "public_key": "pub1",
                    "encrypted": "enc1",
                },
                b"pub1",
            ),
            (
                "1" * 64,
                {
                    "storage": ProtectedKeyStorage.NACL,
                    "public_key": "pub2",
                    "encrypted": "enc2",
                },
                b"pub2",
            ),
        ]

        executor = MigrationExecutor(connection)
        executor.migrate(migrate_from)
        old_apps = executor.loader.project_state(migrate_from).apps
        for old_fingerprint, old_private_key, old_public_key in old_keys:
            old_apps.get_model("signing", "Key").objects.create(
                purpose=KeyPurpose.UEFI,
                fingerprint=old_fingerprint,
                private_key=old_private_key,
                public_key=old_public_key,
            )

        executor.loader.build_graph()
        executor.migrate(migrate_to)
        new_apps = executor.loader.project_state(migrate_to).apps

        for new_fingerprint, new_private_key, new_public_key in new_keys:
            key = new_apps.get_model("signing", "Key").objects.get(
                purpose=KeyPurpose.UEFI, fingerprint=new_fingerprint
            )
            self.assertEqual(key.private_key, new_private_key)
            self.assertEqual(bytes(key.public_key), new_public_key)

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        for old_fingerprint, old_private_key, old_public_key in old_keys:
            key = old_apps.get_model("signing", "Key").objects.get(
                purpose=KeyPurpose.UEFI, fingerprint=old_fingerprint
            )
            self.assertEqual(key.private_key, old_private_key)
            self.assertEqual(bytes(key.public_key), old_public_key)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The collection manager for debian:qa-results collections."""

import datetime as dt
import re
from typing import Any, override

from django.db import IntegrityError
from django.db.models import Q, Window
from django.db.models.fields.json import KT
from django.db.models.functions import Rank
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianQAResult,
)
from debusine.db.models import Artifact, CollectionItem, User, WorkRequest
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)


class DebianQAResultsManager(CollectionManagerInterface):
    """Manage collection of category debian:qa-results."""

    COLLECTION_CATEGORY = CollectionCategory.QA_RESULTS
    VALID_BARE_DATA_CATEGORIES = frozenset({BareDataCategory.QA_RESULT})
    VALID_ARTIFACT_CATEGORIES = frozenset(
        {
            ArtifactCategory.AUTOPKGTEST,
            ArtifactCategory.BLHC,
            ArtifactCategory.LINTIAN,
            ArtifactCategory.PIUPARTS,
        }
    )

    def _expire(
        self,
        *,
        user: User,
        workflow: WorkRequest | None,
        exclude: set[CollectionItem],
    ) -> None:
        """Expire old items from the collection."""
        # For each task/package/architecture combination, keep a configured
        # number of most recent entries.  Clean up other matching items.
        old_items = self.collection.child_items.active()
        rank = Window(
            Rank(),
            partition_by=(
                KT("data__task_name"),
                KT("data__package"),
                KT("data__architecture"),
            ),
            order_by="-data__timestamp",
        )
        most_recent = old_items.annotate(rank=rank).filter(
            rank__lte=self.collection.data.get("old_items_to_keep", 5)
        )
        old_items.exclude(id__in={item.id for item in exclude}).exclude(
            id__in=most_recent.values("id")
        ).update(
            removed_by_user=user,
            removed_by_workflow=workflow,
            removed_at=timezone.now(),
        )

    @override
    def do_add_bare_data(
        self,
        category: BareDataCategory,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        data: dict[str, Any] | None = None,
        name: str | None = None,  # noqa: U100
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        if data is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.QA_RESULTS} requires data"
            )
        if name is not None:
            raise ItemAdditionError(
                f"Cannot use an explicit item name when adding to "
                f"{CollectionCategory.QA_RESULTS}"
            )

        package = data["package"]
        version = data["version"]
        architecture = data["architecture"]
        timestamp = data["timestamp"]
        work_request_id = data["work_request_id"]
        work_request = WorkRequest.objects.get(id=work_request_id)

        qa_result = DebianQAResult(
            task_name=work_request.task_name,
            package=package,
            version=version,
            architecture=architecture,
            timestamp=timestamp,
            work_request_id=work_request_id,
            result=WorkRequest.Results(work_request.result),
        )

        name_elements = [
            work_request.task_name,
            package,
            version,
            architecture,
            str(work_request_id),
        ]
        name = "_".join(name_elements)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.BARE],
                user=user,
                workflow=workflow,
            )

        try:
            item = CollectionItem.objects.create_from_bare_data(
                category,
                parent_collection=self.collection,
                name=name,
                data=qa_result.model_dump(mode="json"),
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

        # Run the expiry logic over the whole collection, by way of a
        # crash-only design: if something went wrong in the past, or if we
        # change the expiry rules, this will catch up.  If we need to
        # optimize this later, adding some good indexes (with the aid of
        # production-scale data) will probably be enough.
        self._expire(user=user, workflow=workflow, exclude={item})

        return item

    @override
    def do_add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,  # noqa: U100
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        if variables is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.QA_RESULTS} requires variables"
            )
        if name is not None:
            raise ItemAdditionError(
                f"Cannot use an explicit item name when adding to "
                f"{CollectionCategory.QA_RESULTS}"
            )

        package = variables["package"]
        version = variables["version"]
        architecture = variables["architecture"]
        timestamp = variables["timestamp"]
        work_request_id = variables["work_request_id"]
        work_request = WorkRequest.objects.get(id=work_request_id)

        qa_result = DebianQAResult(
            task_name=work_request.task_name,
            package=package,
            version=version,
            architecture=architecture,
            timestamp=timestamp,
            work_request_id=work_request_id,
            result=WorkRequest.Results(work_request.result),
        )

        name_elements = [
            work_request.task_name,
            package,
            version,
            architecture,
            str(work_request_id),
        ]
        name = "_".join(name_elements)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[
                    CollectionItem.Types.BARE,
                    CollectionItem.Types.ARTIFACT,
                ],
                user=user,
                workflow=workflow,
            )

        try:
            item = CollectionItem.objects.create_from_artifact(
                artifact,
                parent_collection=self.collection,
                name=name,
                data=qa_result.model_dump(mode="json"),
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

        self._expire(user=user, workflow=workflow, exclude={item})

        return item

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()

    @override
    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem based on the query.

        :param query: `latest:TASKNAME_PACKAGE_ARCHITECTURE`.  If more than
          one possible CollectionItem matches the query: return the one with
          the highest timestamp, tie-breaking on the one that was most
          recently added to the collection.
        """
        # parent_category is technically redundant, but helps PostgreSQL to
        # use the correct index: db_ci_qa_results_latest_idx.
        query_filter = Q(parent_category=CollectionCategory.QA_RESULTS)
        order_by = ["created_at"]

        if m := re.match(r"^latest:([^_]+)_([^_]+)_([^_]+)$", query):
            query_filter &= Q(
                data__task_name=m.group(1),
                data__package=m.group(2),
                data__architecture=m.group(3),
            )
            order_by.insert(0, "data__timestamp")
        else:
            raise LookupError(f'Unexpected lookup format: "{query}"')

        try:
            return (
                self.collection.child_items.active()
                .filter(query_filter)
                .latest(*order_by)
            )
        except CollectionItem.DoesNotExist:
            return None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for exceptions."""

import json
import logging
import unittest
from unittest import mock

from django.test import RequestFactory
from rest_framework import status

from debusine.server.exceptions import (
    DebusineAPIException,
    debusine_exception_handler,
)
from debusine.server.views import ProblemResponse
from debusine.test.django import TestCase


class DebusineAPIExceptionTests(unittest.TestCase):
    """Tests for DebusineAPIException."""

    def test_simple(self) -> None:
        e = DebusineAPIException(
            title="Title",
            detail="Detail",
            status_code=418,
            rollback_transaction=False,
            validation_errors={"foo": ["bar"]},
        )
        self.assertEqual(e.debusine_title, "Title")
        self.assertEqual(e.debusine_detail, "Detail")
        self.assertEqual(e.debusine_status_code, 418)
        self.assertEqual(e.debusine_validation_errors, {"foo": ["bar"]})
        self.assertEqual(e.rollback_transaction, False)

    def test_pydantic_validation_errors(self) -> None:
        e = DebusineAPIException(
            title="Title",
            pydantic_validation_errors=[
                {"loc": ("foo",), "msg": "bar", "type": "baz", "input": ""}
            ],
        )
        self.assertEqual(e.debusine_validation_errors, {"foo": ["bar"]})

    def test_both_validation_errors(self) -> None:
        with self.assertRaises(
            ValueError,
            msg=(
                "Cannot accept both validation_errors and "
                "pydantic_validation_errors."
            ),
        ):
            DebusineAPIException(
                title="Title",
                validation_errors={},
                pydantic_validation_errors=[],
            )


class DebusineExceptionHandler(TestCase):
    """Tests for debusine_exception_handler."""

    def patch_exception_handler(
        self, return_value: ProblemResponse | None
    ) -> None:
        """django_rest exception_handler() return return_value."""
        exception_handler_patcher = mock.patch(
            "debusine.server.exceptions.exception_handler"
        )
        exception_handler_mocked = exception_handler_patcher.start()
        exception_handler_mocked.return_value = return_value
        self.addCleanup(exception_handler_patcher.stop)

    def test_return_exception_handler_status_code(self) -> None:
        """debusine_exception_handler() return status_code from rest()."""
        expected_status_code = status.HTTP_418_IM_A_TEAPOT

        self.patch_exception_handler(
            mock.create_autospec(
                spec=ProblemResponse, status_code=expected_status_code
            )
        )

        response = debusine_exception_handler(Exception(), {})

        self.assertEqual(response.status_code, expected_status_code)

    def test_return_detail_from_exc(self) -> None:
        """debusine_exception_handler() return detail from exception."""
        expected_detail = "This is the detail"

        exc_mock = mock.create_autospec(
            spec=ProblemResponse, detail=expected_detail
        )

        response = debusine_exception_handler(exc_mock, {})

        content = json.loads(response.content)

        self.assertEqual(content["title"], "Error")
        self.assertEqual(content["detail"], expected_detail)

    def test_return_400_status_code(self) -> None:
        """debusine_exception_handler() return default HTTP 400 status code."""
        self.patch_exception_handler(None)

        response = debusine_exception_handler(Exception(), {})

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_does_not_return_detail(self) -> None:
        """debusine_exception_handler does not return detail."""
        self.patch_exception_handler(None)

        response = debusine_exception_handler(Exception(), {})

        content = json.loads(response.content)

        self.assertNotIn("detail", content)

    def test_logs_error_logs_detail(self) -> None:
        """debusine_exception_handler logs error with no relevant detail."""
        expected_detail = "Detail to be logged"

        exc_mock = mock.create_autospec(
            spec=ProblemResponse, detail=expected_detail
        )

        traceback_formatted = "This is the traceback"
        patcher = mock.patch("traceback.format_exc")
        format_exc_mock = patcher.start()
        format_exc_mock.return_value = traceback_formatted

        self.addCleanup(patcher.stop)

        debusine_exception_handler(exc_mock, {})

        with self.assertLogsContains(
            f"ERROR:debusine.server.exceptions:Server exception. "
            f"status_code: 400 detail: {expected_detail} "
            f"traceback: {traceback_formatted}",
            logger="debusine",
            level=logging.ERROR,
        ):
            debusine_exception_handler(exc_mock, {})

    def test_debusineapiexception(self) -> None:
        """Test turning DebusineAPIException into ProblemResponse."""
        exc = DebusineAPIException(
            title="title",
            detail="detail",
            validation_errors={"validation": "errors"},
            status_code=status.HTTP_418_IM_A_TEAPOT,
        )
        with self.assertLogsContains(
            "I'm a Teapot: None", logger="django.request", level=logging.INFO
        ):
            response = debusine_exception_handler(exc, {})
        self.assertIsInstance(response, ProblemResponse)
        self.assertEqual(response.status_code, status.HTTP_418_IM_A_TEAPOT)
        self.assertEqual(
            response.headers["Content-Type"], "application/problem+json"
        )
        self.assertEqual(
            json.loads(response.content),
            {
                'detail': 'detail',
                'title': 'title',
                'validation_errors': {'validation': 'errors'},
            },
        )
        self.assertTrue(getattr(response, "_has_been_logged"))

    def test_debusineapiexception_with_request(self) -> None:
        exc = DebusineAPIException(
            title="title",
            detail="detail",
            validation_errors={"validation": "errors"},
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
        )
        with self.assertLogsContains(
            "Service Unavailable: /path/",
            logger="django.request",
            level=logging.INFO,
        ):
            response = debusine_exception_handler(
                exc, {"request": RequestFactory().get("/path/")}
            )
        self.assertIsInstance(response, ProblemResponse)
        self.assertEqual(
            response.status_code, status.HTTP_503_SERVICE_UNAVAILABLE
        )
        self.assertEqual(
            response.headers["Content-Type"], "application/problem+json"
        )
        self.assertEqual(
            json.loads(response.content),
            {
                'detail': 'detail',
                'title': 'title',
                'validation_errors': {'validation': 'errors'},
            },
        )
        self.assertTrue(getattr(response, "_has_been_logged"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage workers."""

import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Any, NoReturn, cast, override

from django.core.management import CommandError, CommandParser
from django.db import transaction

from debusine.db.models import Token, WorkRequest, Worker
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import Workers
from debusine.tasks.models import WorkerType
from debusine.utils.input import EditorInteractionError, InvalidData, YamlEditor


class Command(DebusineBaseCommand):
    """Command to manage workers."""

    help = "Manage workers."

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        create = subparsers.add_parser(
            "create", help="Create a worker and print its activation token."
        )
        create.add_argument(
            "--worker-type",
            choices=["external", "signing"],
            type=WorkerType,
            default=WorkerType.EXTERNAL,
        )
        create.add_argument(
            "fqdn", help="Fully-qualified domain name of the worker"
        )

        enable = subparsers.add_parser("enable", help="Enable a worker.")
        enable.add_argument(
            "--worker-type",
            choices=["external", "signing"],
            type=WorkerType,
            default=WorkerType.EXTERNAL,
        )
        enable.add_argument("worker", help="Name of the worker to enable")

        disable = subparsers.add_parser(
            "disable",
            help=(
                "Disable a worker. This disables the worker's token (so the "
                "worker cannot communicate with the server) and de-assigns "
                "work requests for the worker, which will be assigned to "
                "another worker. No attempt is made to stop any worker's "
                "operations, but the server will not accept any results."
            ),
        )
        disable.add_argument(
            "--worker-type",
            choices=["external", "signing"],
            type=WorkerType,
            default=WorkerType.EXTERNAL,
        )
        disable.add_argument("worker", help="Name of the worker to disable")

        edit_metadata = subparsers.add_parser(
            "edit_metadata", help="Edit a worker's metadata."
        )
        edit_metadata.add_argument(
            "--set",
            type=Path,
            help=(
                "Filename with the metadata in YAML"
                " (note: its contents will be displayed to users in the web UI)"
            ),
            metavar="PATH",
        )
        edit_metadata.add_argument(
            "worker", help="Name of the worker whose metadata should be edited"
        )

        list_ = subparsers.add_parser("list", help="List workers.")
        list_.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )

    def get_worker(self, name_or_token: str, worker_type: WorkerType) -> Worker:
        """Get a worker by name or token, checking its type."""
        worker = Worker.objects.get_worker_or_none(name_or_token)
        if worker is None:
            worker = Worker.objects.get_worker_by_token_key_or_none(
                name_or_token
            )

        if worker is None:
            raise CommandError("Worker not found", returncode=3)
        if worker.worker_type != worker_type:
            raise CommandError(
                f'Worker "{worker.name}" is of type "{worker.worker_type}", '
                f'not "{worker_type}"',
                returncode=4,
            )

        return worker

    def handle_create(self, *args: Any, **options: Any) -> NoReturn:
        """Create a worker and print its activation token."""
        with transaction.atomic():
            activation_token = Token.objects.create_worker_activation()
            Worker.objects.create_with_fqdn(
                options["fqdn"],
                activation_token=activation_token,
                worker_type=options["worker_type"],
            )
            print(activation_token.key)

        raise SystemExit(0)

    def handle_enable(self, *args: Any, **options: Any) -> NoReturn:
        """Enable a worker."""
        worker = self.get_worker(options["worker"], options["worker_type"])
        # By this point the worker cannot be a Celery worker, and a database
        # constraint ensures that all non-Celery workers have a token.
        assert worker.token is not None

        worker.token.enable()

        raise SystemExit(0)

    def handle_disable(self, *args: Any, **options: Any) -> NoReturn:
        """Disable a worker."""
        worker = self.get_worker(options["worker"], options["worker_type"])
        # By this point the worker cannot be a Celery worker, and a database
        # constraint ensures that all non-Celery workers have a token.
        assert worker.token is not None

        with transaction.atomic():
            worker.token.disable()
            for work_request in worker.assigned_work_requests.filter(
                status__in={
                    WorkRequest.Statuses.PENDING,
                    WorkRequest.Statuses.RUNNING,
                }
            ):
                work_request.de_assign_worker()

        raise SystemExit(0)

    def _set_metadata_from_file(self, worker: Worker, path: Path) -> bool:
        """Set the worker's static metadata from a YAML file."""
        editor: YamlEditor[dict[str, Any]] = YamlEditor({})
        try:
            shutil.copy2(path, editor.editor_file)
            editor.read_edited()
        except (OSError, InvalidData, EditorInteractionError) as e:
            self.stderr.write("Error: " + str(e))
            return False
        finally:
            editor.cleanup()

        if editor.value:
            worker.static_metadata = editor.value
            worker.save()
            self.stdout.write(f"debusine: metadata set for {worker.name}")
            return True
        else:
            self.stderr.write("Error: file is empty")
            return False

    def _edit_metadata(self, worker: Worker) -> bool:
        editor: YamlEditor[dict[str, Any]] = YamlEditor(worker.static_metadata)
        if editor.edit():
            worker.static_metadata = editor.value
            worker.save()
            self.stdout.write(f"debusine: metadata set for {worker.name}")
            return True
        else:
            return False

    def handle_edit_metadata(self, *args: Any, **options: Any) -> NoReturn:
        """Edit a worker's metadata."""
        worker_name = options["worker"]

        try:
            worker = Worker.objects.get(name=worker_name)
        except Worker.DoesNotExist:
            raise CommandError(
                f'Error: worker "{worker_name}" is not registered\n'
                f'Use the command `debusine-admin worker list` to list the '
                f'existing workers',
                returncode=3,
            )

        if options["set"]:
            if self._set_metadata_from_file(worker, options["set"]):
                raise SystemExit(0)
            else:
                raise SystemExit(3)
        else:
            if self._edit_metadata(worker):
                raise SystemExit(0)
            else:
                raise SystemExit(3)

    def handle_list(self, *args: Any, **options: Any) -> NoReturn:
        """List the workers."""
        workers = Worker.objects.all()
        Workers(options["yaml"]).print(workers, self.stdout)
        raise SystemExit(0)

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        func(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The collection manager for debian:environments collections."""

import datetime as dt
import re
from typing import Any, override

from django.db import IntegrityError
from django.db.models import Q
from django.utils import timezone

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import Artifact, CollectionItem, User, WorkRequest
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)


class DebianEnvironmentsManager(CollectionManagerInterface):
    """Manage collection of category debian:environments."""

    COLLECTION_CATEGORY = CollectionCategory.ENVIRONMENTS
    VALID_ARTIFACT_CATEGORIES = frozenset(
        {
            ArtifactCategory.SYSTEM_TARBALL,
            ArtifactCategory.SYSTEM_IMAGE,
        }
    )

    @override
    def do_add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,  # noqa: U100
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add the artifact into the managed collection.

        :param artifact: artifact to add
        :param user: user adding the artifact to the collection
        :param workflow: workflow adding the artifact to the collection
        :param variables: may include `codename` to set the codename of the
          distribution version, `variant` to indicate what kind of tarball
          or image this is, or `backend` to indicate which backend this is
          intended for
        :param replace: if True, replace an existing item with the same name
        """
        artifact_data = artifact.data

        if variables is not None and "codename" in variables:
            codename = variables["codename"]
        else:
            codename = artifact_data["codename"]
        if variables is not None and "variant" in variables:
            variant = variables["variant"]
        else:
            variant = None
        if variables is not None and "backend" in variables:
            backend = variables["backend"]
        else:
            backend = None
        architecture = artifact_data["architecture"]

        if artifact.category == ArtifactCategory.SYSTEM_TARBALL:
            format_ = "tarball"
        else:
            assert artifact.category == ArtifactCategory.SYSTEM_IMAGE
            format_ = "image"

        data = {
            "codename": codename,
            "architecture": architecture,
            "variant": variant,
            "backend": backend,
        }

        name_elements = [format_, codename, architecture]
        if variant is not None or backend is not None:
            name_elements.append(variant or "")
            name_elements.append(backend or "")
        name = ":".join(name_elements)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.ARTIFACT],
                user=user,
                workflow=workflow,
            )

        try:
            return CollectionItem.objects.create_from_artifact(
                artifact,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        # Retention logic will remove the item, now only marked as removed.
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()

    @override
    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem based on the query.

        :param query: Examples "match:codename=CODENAME:architecture=ARCH",
          "match:format=tarball:architecture=ARCH:variant=VARIANT". If more
          than one possible CollectionItem matches the query: return the
          most recently added one.
        """
        if m := re.match(r"^match:(.+)$", query):
            filters_string = m.group(1)
        else:
            raise LookupError(f'Unexpected lookup format: "{query}"')

        filters = dict(item.split("=", 1) for item in filters_string.split(":"))
        query_filter = Q(child_type=CollectionItem.Types.ARTIFACT)

        if filters.get("format") == "tarball":
            query_filter &= Q(category=ArtifactCategory.SYSTEM_TARBALL)
        elif filters.get("format") == "image":
            query_filter &= Q(category=ArtifactCategory.SYSTEM_IMAGE)

        if "codename" in filters:
            query_filter &= Q(data__codename=filters["codename"])
        if "architecture" in filters:
            query_filter &= Q(data__architecture=filters["architecture"])
        if "variant" in filters:
            query_filter &= Q(data__variant=filters["variant"] or None)
        if "backend" in filters:
            query_filter &= Q(data__backend=filters["backend"])

        try:
            return (
                self.collection.child_items.active()
                .filter(query_filter)
                .latest("created_at")
            )
        except CollectionItem.DoesNotExist:
            return None
